package fullstack

import (
	"testing"
	"time"

	xbus "bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"github.com/stretchr/testify/assert"
)

func TestEndToEndMultiConsumer(t *testing.T) {
	for _, tt := range []struct {
		name  string
		graph string
	}{
		{"By Role", `
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> target
`},
		{"By actor name", `
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: target
  type: consumer
  actors: [jojo, jojo2]
  inputs: [default]
edges:
- source -> target
`},
	} {
		t.Run(tt.name, func(t *testing.T) {
			ft := InitFullstack(t, true)
			if ft == nil {
				return
			}
			defer ft.Shutdown()

			ft.CreateRole(api.Actor_EMITTER, "anyemitter")
			_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

			assert.Equal(t, "bob", emitter.Name)

			hit := make(chan *api.Envelope, 1)

			ft.CreateRole(api.Actor_CONSUMER, "anyconsumer")

			_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
			xsub, err := consumer.Subscribe(
				func(apc *xbus.ActorProcessingContext) error {
					t.Logf("Here i am")
					r, err := apc.ReadEnvelope("default")
					if err != nil {
						ft.T.Error(err)
						return err
					}
					e, err := r.Complete(time.Second)
					if err != nil {
						ft.T.Error(err)
						return err
					}
					hit <- e
					return nil
				})
			assert.Nil(t, err)
			defer xsub.Unsubscribe()

			assert.Equal(t, "jojo", consumer.Name)

			_, consumer2 := ft.CreateConsumer("jojo2", []string{"anyconsumer"})
			xsub2, err := consumer2.Subscribe(
				func(apc *xbus.ActorProcessingContext) error {
					t.Logf("Here i am")
					r, err := apc.ReadEnvelope("default")
					if err != nil {
						ft.T.Error(err)
						return err
					}
					e, err := r.Complete(time.Second)
					if err != nil {
						ft.T.Error(err)
						return err
					}
					hit <- e
					return nil
				})
			assert.Nil(t, err)
			defer xsub2.Unsubscribe()

			ft.InstallGraph(tt.graph)

			env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

			envStateChan, sub := ft.GetEnvStateChan(emitter, api.UUID{})
			defer sub.Unsubscribe()

			status, err := emitter.SendEnvelope(nil, api.DefaultOutput, env)
			assert.Nil(t, err)
			assert.Equal(t, api.EnvelopeAck_ACCEPTED, status)

			timeout := time.After(5 * time.Second)
			select {
			case <-timeout:
				t.Errorf("Process state not updated")
			case envState := <-envStateChan:
				assert.Equal(t, api.Process_RUNNING, envState.Status)
			}

			timeout = time.After(5 * time.Second)
			select {
			case <-timeout:
				t.Errorf("No envelope received")
			case lastEnvelope := <-hit:
				assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
			}

			timeout = time.After(5 * time.Second)
			select {
			case <-timeout:
				t.Errorf("Second envelope not received")
			case lastEnvelope := <-hit:
				assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
			}

			timeout = time.After(5 * time.Second)
			select {
			case <-timeout:
				t.Errorf("Process state not updated to 'done'")
			case envState := <-envStateChan:
				assert.Equal(t, api.Process_DONE, envState.Status)
			}
		})
	}
}
